;       Last change    August 1993
(in-package lib)

;; Help functions
;; Help information is kept in files under the directory help, and
;; end with the suffix .hlp
;; A file exists for each command, for general help, and hopefully for
;; topics (like syntax) later sometime too.

(defconstant *eof*  '|eof|
     "Returned by read functions to signal end of file")
(defvar *help-path* 
#+lucid "epilog:help;.hlp"
#-lucid "epilog:help;<help>.hlp"
     "Path indicating where to find help")

; help gets the general help file with a list of commands if no
; items are given to it; otherwise it gets specific help for those items

(defun help (&rest items)
  (if items
      (mapcar 'get-help items)
      (get-help 'help))
  NIL)

(defun ? (&rest items) (apply 'help items))

; get help determines the name of the file help will be in, gets the file,
; and prints the information in it
; If no file exists a message is printed.

(defun get-help (item &aux
		      (helpfile (merge-pathnames 
				 (string-downcase (symbol-name item))
						 *help-path*)))
  (with-open-file (ifile helpfile :if-does-not-exist nil)
		  (cond
		   (ifile
		    (print-line)
		    (do ((line (read-line ifile nil *eof*)
			       (read-line ifile nil *eof*)))
			((eq line *eof*) nil)
			(print-line line)))
		   (t (no-help item))
		   ))
  nil
  )

; no-help prints a message saying that help is not available for the
; given item

(defun no-help (item)
  (print-line "No help available for " item)
  nil)
