;       Last change    August 1993
(in-package epilog)

(defun lookup (phi &optional proof-prop save-classes assumptions &aux prob res
		   classes-to-check) 
  ;; simplest check - after normalizing we already have it
  (if (and (setq prob (getp phi 'prob))
	   (check-proof-prop phi proof-prop)
	   (if (getp phi 'assumption) (member phi assumptions) t)
	   (not (and (< (getp phi 'c-prob) 1) 
		     (or (getp phi 'contradict) (getp phi 'support))))
	   )
      (list 'YES (list phi) prob)
    
    (if (and (getp phi 'negated) (equal (getp phi 'pred) 'epilib::true))
	(let ((res (car (lookup (epilib::norm-wff (list (car (getp phi 'arguments)) 'epilib::true))))))
	  (case (car res) 
	    (YES (list 'NO (cadr res) (caddr res)))
	    (NO nil)
	    (otherwise nil)
	    ))

      ;; now look up other formulas under this ones classifications
      ;; to see if there is an equivalent formula
      (let ((class (sort-classes (classify phi)))
	    (ttype (getp phi 'wff-type))
	    (p1 (getp phi 'pred))
	    (a1 (getp phi 'arguments))
	    (neg (neg-p phi))
	    (condit (getp phi 'condtional))
	    a2 p2)
	
	(if save-classes 
	    (progn (setq *save-class* class)
		   (setq *save-intermediate* *intermediate-class*))
	  (remprop phi 'lit-classes))
	
	(setq classes-to-check (choose-classes class))
	;; just check one class
	;; unless equivalent classes - check them also!
	(dolist (w (apply 'append
			  (mapcar #'(lambda (a)
				      (get-class a))
				  classes-to-check)))
	  (when (and (check-proof-prop w proof-prop)
		     (> (getp w 'c-prob) 0)
		     (if (getp w 'assumption)
			 (member w assumptions) t))
	    (cond ((equal w phi) 
		   (return-from lookup 
		     (list 'YES (list w) (getp w 'prob))))
		  ((not (equal (getp w 'wff-type) ttype)) nil)
		  ((not (equal (getp w 'conditional) condit)) nil)
		  (condit 
		   (setq res (compare-conditionals phi w))
		   (if res (return-from lookup res)))
		  ((not (and (setq p2 (getp w 'pred))
			     (setq a2 (getp w 'arguments))
			     a1 p1
			     (lookup-arglists a1 a2 p1 '/)))
		   nil)
		  ((compare-predicates neg p1 (neg-p w) p2 t nil)
		   (return-from lookup (list 'YES (list w) (getp w 'PROB))))
		  ((compare-predicates neg p1 (neg-p w) p2 nil nil)
		   (return-from lookup (list 'NO (list w) (getp w 'PROB))))
		  )
	    )
	  )
	) )
    )
  )

(defun choose-classes (classes &aux
			       newcl lastt)
  (dolist (c classes)
	  (cond ((listp (car c)) (setq newcl (append newcl (list c))))
		((or (null lastt) (equal (cadr c) lastt))
		 (setq newcl (append newcl (list c))))
		(t (return-from choose-classes newcl))
		)
	  (setq lastt (cadr c))
	  )
  newcl)


(defun check-proof-prop (phi prop)
  (if prop
      (cond ((getp phi prop) t)
	    ((getp phi (if (equal prop 'PROOF) 'DISPROOF 'PROOF)) nil)
	    (t t))
      t)
  )

(defun compare-conditionals (wff1 wff2)
  (declare (ignore wff1 wff2))
  )

; be careful here - specs add the equivalence info but don't take into
; consideration modal embedding - how to handle that?
(defun lookup-arglists (arglist1 arglist2 pred &optional (subnet '/)
				 &aux n-subnet a1 a2)
  ; if this isn't a conditional, then there should be no variables?
  ; if no variables, then the only thing allowed to differ are predicates
	
  (when (subnet-pred pred)
	(setq a1 (remove (last1 arglist1) arglist1))
	(setq a2 (remove (last1 arglist2) arglist2))
	(when (lookup-arglists a1 a2 nil subnet)
	      (setq n-subnet (build-symbol (append (list subnet) a1
		  (list '-) (list pred))))
	      (return-from lookup-arglists 
			   (same-term (last1 arglist1) (last1 arglist2) nil n-subnet))
	      )
	(return-from lookup-arglists nil))

  (and (equal (length arglist1) (length arglist2))
       (eval (append (list 'and) 
		     (mapcar #'(lambda (a b) 
				       (if (same-term a b pred subnet) t nil))
			     arglist1 arglist2))) )
  )

(defun same-term (a b pred subnet &aux atype attype)
  (cond ((equal a b) t)
	((member a (equivalent-items b subnet)) t)
	((not (equal (setq atype (entity-type a)) (entity-type b))) nil)
	((equal atype 'term)
	 (if (equal (setq attype (term-type a)) (term-type b))
	     (case attype
		   (wff (lookup-same-wff a b pred subnet))
		   ((modified function)
		    (and (equal (getp a 'operator) (getp b 'operator))
			 (equal (getp a 'negated) (getp b 'negated))
			 (lookup-arglists (getp a 'arguments) 
					  (getp b 'arguments) nil subnet)))
		   (otherwise nil))
	     ))
	((equal atype 'wff) (lookup-same-wff a b pred subnet))
	(t nil))
  )

(defun lookup-same-wff (a b pred subnet &aux atype)
  (when (equal (setq atype (wff-type a)) (wff-type b))
	(case atype
	      ((prefix causal)
	       (and (equal (getp a 'pred) (getp b 'pred))
		    (equal (getp a 'negated) (getp b 'negated))
		    (lookup-arglists (getp a 'arguments) (getp b 'arguments)
				     (getp a 'pred) subnet)))
	      (episodic
	       (let ((argsa (getp a 'arguments))
		     (argsb (getp b 'arguments)))
		 (and (equal (getp a 'pred) (getp b 'pred)) ; compare
		      (equal (getp a 'negated) (getp b 'negated))
		      (same-term (last1 argsa) (last1 argsb) pred subnet)
		      (lookup-same-wff (car argsa) (car argsb) pred subnet))) )
	      (otherwise nil))
	)
  )

(defun verify (phi &optional proof-prop (level 2) (save-classes nil) assumptions
	       &aux res ev)
  (cond ((member phi '(YES NO))
	 (list phi nil 1))
	; lookup does specialist tests for predicates
	((setq res (lookup phi proof-prop 
			   (if (or (getp phi '+lits) (getp phi '-lits)) save-classes)
			   assumptions))
	 (if (and (equal level 1) (not (= (calc-prob-list (caddr res)) 1)))
	     nil
	   res))
	(t
	 (case (getp phi 'wff-type)
	   ((prefix episodic)		; use specialists for whole literal
	    (when (and *spec-evaluate*
		       (setq ev (spec-evaluate phi *specialist-eval-effort* assumptions))
		       (not (equal ev 'UNKNOWN)))
	      (setp phi 'spec-eval ev)
	      (let ((swff (if (equal ev 'NO) 
			      (normalize (negate (getp phi 'print)) '+ nil nil)
			    phi)))
		(change-property swff 'spec-eval ev)
		(list ev (list swff) 1))) )
					; most should have been handled in lookup
	   (causal nil)
	   ((quantified logical)
	    (let* ((+keys (getp phi '+keys))
		   (-keys (getp phi '-keys))
		   (+other (getp phi '+other))
		   (-other (getp phi '-other))
		   res subs newphi par newp change)
	      (dolist (k (append -keys +keys +other -other))
		(unless (equal k phi)
		  (setq res (verify k proof-prop level nil
				    assumptions))
		  (when res
		    (setq subs (append subs 
				       (list (list k (car res)))))
		    (setq newp 
		      (combine-probs newp (caddr res)))
		    (setq par (union par (cadr res)))
		    (setq change t))
		  )
		)
	      (when change
		(setq newphi (normalize 
			      (substitute-prop phi subs) '+ nil nil))
		(when proof-prop (setp newphi proof-prop t))
		(setq newp (combine-probs newp *subst-prob*))
		(setq newp (combine-probs newp *normed-prob*))
		(list newphi par newp)
		))
	    )
	   
	   (modified nil)
	   (constant nil)
	   (variable nil) 
	   )
	 ))
  )

